package com.gxzc.zen.orm.aop

import com.gxzc.zen.common.util.SpringContextHolder
import com.gxzc.zen.orm.annotation.MultiTransactional
import org.aspectj.lang.ProceedingJoinPoint
import org.aspectj.lang.annotation.Around
import org.aspectj.lang.annotation.Aspect
import org.aspectj.lang.reflect.MethodSignature
import org.slf4j.LoggerFactory
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty
import org.springframework.stereotype.Component
import org.springframework.transaction.jta.JtaTransactionManager
import kotlin.reflect.KClass

/**
 *
 * @author NorthLan at 2018/2/3
 */
@Aspect
@Suppress("unused")
@Component
@ConditionalOnProperty(prefix = "orm", name = ["multi-datasource-enable"], havingValue = "true", matchIfMissing = true)
class MultiDataSourceAspect {
    companion object {
        private val logger = LoggerFactory.getLogger(MultiDataSourceAspect::class.java)
    }

    init {
        logger.info("MultiDataSourceAspect initializing...")
    }

    /**
     * 一致性事务切面解决方案
     */
    @Around("@annotation(com.gxzc.zen.orm.annotation.MultiTransactional)")
    fun multiTransactionAround(joinPoint: ProceedingJoinPoint): Any? {
        logger.info("@MultiTransactional aspect...")
        val methodName = joinPoint.signature.name
        val parameterTypes = (joinPoint.signature as MethodSignature).method.parameterTypes
        val method = joinPoint.target::class.java.getMethod(methodName, *parameterTypes)

        if (method.isAnnotationPresent(MultiTransactional::class.java)) {
            val multiTransaction = method.getAnnotation(MultiTransactional::class.java)
            val transactionManager = SpringContextHolder.getBean(JtaTransactionManager::class.java)
            val userTransaction = transactionManager.userTransaction
            return try {
                userTransaction?.begin()
                val r = joinPoint.proceed()
                userTransaction?.commit()
                return r
            } catch (e: Throwable) {
                val rollbackExceptions = multiTransaction.rollbackFor
                val noRollbackExceptions = multiTransaction.noRollbackFor

                val isRollbackExceptionPresent = isPresent(e, rollbackExceptions)
                val isNoRollbackExceptionPresent = isPresent(e, noRollbackExceptions)

                // rollbackFor回滚 noRollbackFor不回滚
                if (isRollbackExceptionPresent || !isNoRollbackExceptionPresent) {
                    userTransaction?.rollback()
                } else {
                    userTransaction?.commit()
                }
            }
        } else {
            return joinPoint.proceed()
        }
    }

    /**
     * 判断指定类型是否存在于注解中
     * @param e 指定抛出类型
     * @param exceptions 注解定义的类型列表
     */
    private fun isPresent(e: Throwable, exceptions: Array<KClass<out Throwable>>): Boolean {
        return exceptions.find {
            e::class.java.isAssignableFrom(it.java) || e::class.java == it::java
        } != null
    }
}